#include "textlib.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

void free_font(typeutil *t)
{
	if (t->face != NULL)
	{
		FT_Done_Face(t->face);
		t->face=NULL;
		t->slot = NULL;
	};
};

void set_font_point_size(typeutil *t, int points)
{
	int error;
	if (t==NULL)
	{
		return;
	};
	if (t->face==NULL)
	{
		return;
	};
	if (points<=0)
	{
		return;
	};
	error = FT_Set_Char_Size(
					t->face,    /* handle to face object           */
					0,       /* char_width in 1/64th of points  */
					points * 64, /* char_height in 1/64th of points */
					0,     /* horizontal device resolution    */
					0 );   /* vertical device resolution      */	
};

void render_char_gray(typeutil *t, char c, char p)
{
	if (t==NULL)
	{
		return;
	};
	if (t->face==NULL)
	{
		return;
	};
	t->glyph_index = FT_Get_Char_Index( t->face, c );
	if (p!=0)
	{
		t->prev_glyph_index = FT_Get_Char_Index( t->face, p );
		FT_Get_Kerning( t->face, 
										t->prev_glyph_index, 
										t->glyph_index,
										ft_kerning_default,
										&t->delta );
		t->pen_x += t->delta.x >> 6;

	};
	t->error = FT_Load_Glyph( t->face, t->glyph_index, FT_LOAD_DEFAULT );
	FT_Get_Glyph(t->slot,&t->glyph);
	FT_Glyph_Get_CBox( t->glyph, ft_glyph_bbox_pixels, &t->bbox );

	FT_Glyph_To_Bitmap( &t->glyph, ft_render_mode_normal, 0, 1);

	t->bitmap = (FT_BitmapGlyph)t->glyph;
	t->pen_x += t->slot->advance.x >> 6;
	t->pen_y += t->slot->advance.y >> 6;   

};

void set_font(typeutil *t, char *filename)
{
	int error;
	if (t==NULL)
	{
		return;
	};
	if (filename==NULL)
	{
		return;
	};
	free_font(t);
	error = FT_New_Face(	t->library,
												filename,
												0,
												&t->face );

	if ( error == FT_Err_Unknown_File_Format )
	{
		printf("Unknown font format for \"%s\"\n",filename);
		t->face=NULL;
		return;
	}
	else if ( error )
	{
		printf("Unknown error reading \"%s\"\n",filename);
		t->face=NULL;
		return;
	}	
	t->slot = t->face->glyph;
};

typeutil	*new_typeutil()
{
	typeutil	*t=NULL;
	int error;
	
	t=(typeutil	*)malloc(sizeof(typeutil));
	if (t==NULL)
	{
		return NULL;
	};

	memset(t,0,sizeof(typeutil));

	error = FT_Init_FreeType( &t->library );
	if (error)
	{
		printf("could not initialize freetype library\n");
		return NULL;
	};

	return t;
};
